from __future__ import print_function

import sys
import json
import logging
import time
try:
    from urlparse import urljoin
except ImportError:
    from urllib.parse import urljoin

import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.exceptions import InsecureRequestWarning
from requests.packages.urllib3.util.retry import Retry


class Salt(object):
    """Salt Rest API Handler"""

    CLIENT_LOCAL = 'local'
    CLIENT_RUNNER = 'runner'
    CLIENT_WHEEL = 'wheel'

    OUTPUT_RAW = 'raw'
    OUTPUT_JSON = 'json'
    OUTPUT_DICT = 'dict'

    def __init__(self, url=None, username=None, passwd=None):
        """
        Salt's Rest API handler

        Args:
            url (str): Salt API web server URL
            username (str): The username to be used to authenticate to the salt-api
            passwd: The password to be used to authenticate to the salt-api
        """
        self.url = url
        self.username = username
        self.password = passwd
        self.token = None
        self.timeout = 60
        self.session = self._new_session()
        self.lookup_interval = 1
        self.max_retries_if_aborted = 3

        self.log = logging.getLogger(__name__)

    def _new_session(self):
        """
        Creates a new requests Session

        Returns:
            request.Session
        """
        session = requests.Session()
        session.verify = False
        retries = Retry(total=3, backoff_factor=5)
        session.mount(self.url, HTTPAdapter(max_retries=retries))
        return session

    def get(self, path, headers=None, timeout=None):
        """
        Executes URL calls using the GET method

        Args:
            path (str): Path to append to the URL
            headers (dict): Extra headers to be passed to the GET call
            timeout (int): The amount of seconds to wait before the request times out

        Returns:
            requests.models.Response
        """
        if not self.session.verify:
            requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

        try:
            ret = self.session.get(
                url=urljoin(self.url, path),
                headers=headers,
                timeout=timeout or self.timeout
            )
            return ret
        except Exception as e:
            self.log.error(e)
            exit(1)

    def post(self, data, path='', headers=None, timeout=None):
        """
        Executes URL calls using the POST method

        Args:
            data (dict): The data to be passed to the POST call
            path (str): The path to be appended to the URL
            headers (dict): Extra headers to be passed to the POST call
            timeout (int): The amount of seconds to wait before the request times out

        Returns:
            requests.models.Response
        """
        if not self.session.verify:
            requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

        retries = self.max_retries_if_aborted or 1

        while True:
            try:
                ret = self.session.post(
                    url=urljoin(self.url, path),
                    data=data,
                    headers=headers,
                    timeout=timeout or self.timeout
                )
                return ret
            except Exception as e:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                self.log.error('Exception type: {}. Exception message: {}'.format(exc_type, exc_value))

                if 'RemoteDisconnected' in str(exc_value) and retries:
                    retries -= 1
                    self.log.warning(
                        'Execution failed because of connection abortion. Remaining retries: {} out of {}...'.format(
                            retries,
                            self.max_retries_if_aborted
                        )
                    )
                    time.sleep(5)
                else:
                    exit(1)

    def login(self):
        """
        Creates an authenticated session with a salt API server

        Returns:

        """
        self.log.debug('Authenticating to salt-api...')
        data = {
            'username': self.username,
            'password': self.password,
            'eauth': 'pam'
        }

        try:
            r = self.post(path='login', data=data)
            self.token = r.json()['return'][0]['token']
            self.session.headers.update({'X-Auth-Token': self.token})
            self.log.debug('Authentication succeed.')
            return r.content
        except Exception:
            self.log.error('Error: Unable to connect to salt-api at `{}`. Return code: {}'.format(r.url, r.status_code))
            exit(1)

    def execute(self, fun, client=None, target=None, args=None, kwargs=None, pillar=None, async=False, async_wait=False,
                output='dict', returner=None):
        """
        Executes a function using the salt API.

        Args:
            fun (str): Name of the function to be executed
            client (str): Type of client to use (local, runner or wheel).
                          If not specified, the `local` client will be used.
            target (str): The minions that should execute this function.
                          If the client is not set to `local`, this parameter will be ignored.
            args (list): List of arguments to be passed to the function.
            kwargs (dict): Dictionary of arguments to be passed to the function.
            pillar (dict): Dictionary of pillar values to be passed to the function. Example: pillar={'sleep': 30}
            async (bool): Whether or not to execute the function asynchronously.
                          If set to `True`, it will return the job ID.
                          If the client is set to `wheel`, this parameter wil be ignored.
            async_wait (bool): If this parameter is set to `True`, it will use async clients, just like the `async`
                               parameter would, but instead of returning the job ID, it will use it to keep pulling
                               the job result until it is completed. It is basically the same as passing `async=True`
                               and calling `Salt.lookup_job` with `until_complete=True`. Crazy right? But super useful
                               for long run functions that you would like to wait for the return, but don't want to
                               be vulnerable to timeouts (You're welcome.).
            output (str): The output format of this method (See output constants specified in this class).
            returner (str): Which salt returner to be passed to the function.

        Returns:
            str, dict: Depends on the format passed to the `output` parameter.
        """

        if not self.token:
            self.login()

        client = client or Salt.CLIENT_LOCAL

        if async or async_wait and not (client == Salt.CLIENT_WHEEL):
            client += '_async'

        data = {
            'client': client,
            'fun': fun,
        }

        if target:
            data.update({'tgt': target})
        if pillar:
            args = args or []
            args.append('pillar={}'.format(json.dumps(pillar)))
        if args:
            data.update({'arg': args})
        if kwargs:
            if client == Salt.CLIENT_LOCAL:
                data.update({'kwarg': kwargs})
            else:
                data.update(**kwargs)

        if returner:
            data.update({'ret': returner})

        self.log.debug('Executing salt command: {}'.format(data))
        ret = self.post(data)
        if async_wait:
            return self.lookup_job(json.loads(ret.content)['return'][0]['jid'], until_complete=True, output=output)
        elif async:
            return json.loads(ret.content)
        elif output == Salt.OUTPUT_DICT:
            return json.loads(ret.content)
        elif output == Salt.OUTPUT_RAW:
            return ret.content
        elif output == Salt.OUTPUT_JSON:
            return str(ret.content)

    def lookup_job(self, jid, until_complete=False, interval=None, output='dict'):
        """
        Retrieves information about a saltstack job.

        Args:
            jid (int): The job ID.
            until_complete (bool): Whether or not to keep pulling the job until it is completed.
            interval (int): Time interval (in seconds) to wait in between checks. If not specified, defaults will be used.
                You can also set the global lookup interval using `self.lookup_interval`.
            output (str): The output format of this method (See output constants specified in this class).
        Returns:
            dict: a dictionary containing the job results
        """
        if not self.token:
            self.login()

        ret = self.execute(client=Salt.CLIENT_RUNNER, fun='jobs.lookup_jid', kwargs={'jid': jid}, output=output)

        if until_complete:
            while ret['return'][0] == {}:
                time.sleep(interval or self.lookup_interval)
                ret = self.lookup_job(jid)

        return ret

    def highstate(self, target, output='dict'):
        """
        Runs highstate on the target minion

        Args:
            target (str): The minion to run the highstate
            output (str): The output format of this method (See output constants specified in this class).
        Returns:
            str, dict: Depends on the format passed to the `output` parameter.
        """
        return self.execute(fun='state.apply', target=target, async_wait=True, output=output)

    def versions(self):
        """
        Returns the versions of salt running on master and its minions

        Returns:
            dict
        """
        ret = self.execute(fun='manage.versions', client=Salt.CLIENT_RUNNER)['return'][0]
        v = {
            'master': ret['Master'],
            'minions': {}
        }

        if 'Up to date' in ret.keys():
            v['minions'].update(ret['Up to date'])

        if 'Minion requires update' in ret.keys():
            v['minions'].update(ret['Minion requires update'])
        return v



from .state import StateOutput


class OrchestrationOutput(StateOutput):
    def __init__(self, ret):
        super(OrchestrationOutput, self).__init__(ret)

    def ordered_result(self, result):
        ordered = {}

        self.log.debug('Ordering orchestration output...')
        for master in result[0]['data']:
            pass




import sys
import colorama
import logging
from collections import OrderedDict
from colorclass import Color, Windows
from terminaltables import AsciiTable, SingleTable

colorama.init()

if sys.stdout.encoding == 'utf-8':
    Windows.enable(auto_colors=True, reset_atexit=True)


class StateOutput:
    def __init__(self, ret):
        """
        Output handler for salt states return objects

        Args:
            ret (dict): The dictionary the is returned as a result of running `state.apply` using the Salt.execute
                        method or any other bubble.Salt method that executes states on minions.
        """
        self.log = logging.getLogger(__name__)
        self.raw_data = ret
        self.data = self.ordered_result(ret)
        self.total_ms = {}
        self.colored = True

        if sys.stdout.encoding == 'utf-8':
            self.safe = False
        else:
            self.safe = True
    
    def ordered_result(self, result):
        """
        Order states by run number

        Args:
            result (dict): The return object of a `state.apply` execution

        Returns:
            OrderedDict
        """
        ordered = {}

        self.log.debug('Ordering state runs...')
        for minions in result['return']:            
            for minion_id in minions.keys():
                states = minions[minion_id]
                try:
                    ordered[minion_id] = OrderedDict(
                        sorted(states.items(), key=lambda k: k[1]['__run_num__']))
                except Exception as e:
                    self.log.error('Error: Unable to sort state results for {} minion'.format(minion_id))
                    self.log.error('State results: {}'.format(states))                    
                    exit(1)

        return ordered

    @staticmethod
    def get_state_name(key):
        """
        Extracts a state name from salt's `state.apply` naming pattern

        Args:
            key (str): The state key from which to extract the state name

        Returns:

        """
        return key.split('_|-')[1]

    def summary(self, max_chars=None):
        """
        Returns a summary of a state run with more meaningful data like `ID`, `duration`, `result` and `changes`.
        ID: the state name extracted from the state key
        Duration: If a state did not run, its duration is set to zero
        Result: `True` or `False` for whether or not the state run successfully
        Changes: `True` or `False` for whether or not the state has made changes to the minion

        Args:            
            max_chars (int): Maximum number of characters to display for state ID.
                             If the ID is greater then `max_chars` ellipsis(...) will be added.
        Returns:
            dict
        """
        ret = {}
        total_ms = {}
        for minion_id in self.data.keys():
            ret[minion_id] = {
                'states': [],
                'total_duration': 0,
                'failed_states': [],
                'raw_states': []
            }

            for state_key, state_data in self.data[minion_id].items():

                ret[minion_id]['total_duration'] += state_data.get('duration', 0)

                state_name = StateOutput.get_state_name(state_key)
                if max_chars:
                    max_chars = abs(max_chars)
                    if len(state_name) > max_chars:
                        state_name = state_name[:max_chars] + '...'

                filtered_data = {
                    'id': state_name,
                    'duration': state_data.get('duration', 0),
                    'result': state_data['result'],
                    'changes': state_data['changes'] != {}
                }
                ret[minion_id]['raw_states'].append(state_data)
                ret[minion_id]['states'].append(filtered_data)

                if not state_data['result']:
                    ret[minion_id]['failed_states'].append(filtered_data)

        return ret

    def tables(self, failed_only=False, max_chars=None, safe=None):
        """
        Creates a list of tables representing the state run for which minion.

        Args:
            failed_only (bool): Whether or not the tables should only contain the failed states
            max_chars (int): Maximum number of characters to display for state ID.
                             If the ID is greater then `max_chars` ellipsis(...) will be added.
            safe (bool): If safe is set to `True`, the table will be created using features that are compatible with
                         most terminals.

        Returns:
            list
        """

        if failed_only:
            states = 'failed_states'
        else:
            states = 'states'

        data = self.summary(max_chars=max_chars)

        if safe is None:
            safe = self.safe

        tables = []
        for minion_id in data:

            if not data[minion_id][states]:
                self.log.debug('No state results found. Skipping `{}`'.format(minion_id))
                continue

            table_data = [['State', 'Time(ms)', 'Result']]

            for state in data[minion_id][states]:
                line = (state['id'], state['duration'], state['result'])
                if state['result']:
                    table_data.append([Color.cyan(item, auto=True) for item in line])
                else:
                    table_data.append([Color.red(item, auto=True) for item in line])

            total_sec = data[minion_id]['total_duration'] / 1000
            if total_sec > 60:
                total_time = '{0:1.2f}min'.format(total_sec / 60)
            else:
                total_time = '{0:1.2f}s'.format(total_sec)
            table_data.append(['Total elapsed time: {}'.format(total_time)])

            if safe:
                table = AsciiTable(table_data)
            else:
                table = SingleTable(table_data)

            table.inner_footing_row_border = True
            table.inner_column_border = False
            table.title = ' {} '.format(minion_id)
            tables.append(table.table)
        return tables

    def graphs(self, failed_only=False, max_chars=None, max_bar_size=30, safe=None):
        """
        Creates a list of tables representing the state run including a more graphical representation of the duration.

        Args:
            failed_only (bool): Whether or not the tables should only contain the failed states
            max_chars (int): Maximum number of characters to display for state ID.
                             If the ID is greater then `max_chars` ellipsis(...) will be added.
            max_bar_size (int): Size of the bar plot equivalent to 100% of the execution time.
            safe (bool): If safe is set to `True`, the table will be created using features that are compatible with
                         most terminals.

        Returns:
            list
        """
        
        if safe is None:
            safe = self.safe

        ticks = ('█', '▌', '|')

        if failed_only:
            states = 'failed_states'
        else:
            states = 'states'

        data = self.summary(max_chars=max_chars)
        tables = []
        if safe:
            the_tick = ticks[2]
        else:
            the_tick = ticks[0]

        for minion_id in data:

            if not data[minion_id][states]:
                self.log.debug('No state results found. Skipping `{}`'.format(minion_id))
                continue

            table_data = [['State', 'Plot', '%', 'ms', 'Result']]
            for state in data[minion_id][states]:
                factor = max_bar_size * state['duration'] / data[minion_id]['total_duration']
                bar = str(the_tick * int(factor))
                percentage = factor * 100 / max_bar_size

                line = (
                    state['id'],
                    bar,
                    '{0:>5.2f}%'.format(percentage),
                    state['duration'],
                    state['result']
                )

                if self.colored:
                    if state['result']:
                        table_data.append([Color.cyan(item, auto=True) for item in line])
                    else:
                        table_data.append([Color.red(item, auto=True) for item in line])
                else:
                    table_data.append([item for item in line])

            total_sec = data[minion_id]['total_duration'] / 1000
            if total_sec > 60:
                total_time = '{0:1.2f}min'.format(total_sec / 60)
            else:
                total_time = '{0:1.2f}s'.format(total_sec)
            table_data.append(['Total elapsed time: {}'.format(total_time)])

            if safe:
                table = AsciiTable(table_data)
            else:
                table = SingleTable(table_data)

            table.title = ' {} '.format(minion_id)
            table.inner_column_border = False
            table.inner_footing_row_border = True
            tables.append(table.table)

        return tables



from setuptools import setup, find_packages
from os import path
from codecs import open

this_folder = path.abspath(path.dirname(__file__))

# Get package version from .version file
with open(path.join(this_folder, '.version')) as f:
    version = f.read()

# Get long description from README.rst file
with open(path.join(this_folder, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='saltypie',
    version=version,
    author='Wilson Santos',
    author_email='wilson@codeminus.org',
    url='https://gitlab.com/cathaldallan/saltypie',
    description='Saltypie - salt-api wrapper',
    long_description=long_description,
    license='MIT',
    keywords='saltstack salt salt-api wrapper',
    classifiers=[
        # 'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3'
    ],
    packages=find_packages(exclude=['docs', 'tests']),
    install_requires=[
        'requests',
        'terminaltables',
        'colorclass',
        'colorama',
        'termcolor',
    ],
)
